@@ -29,6 +29,8 @@ from .transport import ASGIWebSocketAsyncNetworkStream
 JSONMode = typing.Literal["text", "binary"]
 TaskFunction = typing.TypeVar("TaskFunction")
 TaskResult = typing.TypeVar("TaskResult")
+SyncSession = typing.TypeVar("SyncSession", bound="WebSocketSession")
+AsyncSession = typing.TypeVar("AsyncSession", bound="AsyncWebSocketSession")
 
 DEFAULT_MAX_MESSAGE_SIZE_BYTES = 65_536
 DEFAULT_QUEUE_SIZE = 512
@@ -40,6 +42,10 @@ class ShouldClose(Exception):
     pass
 
 
+class EndOfStream(Exception):
+    pass
+
+
 class WebSocketSession:
     """
     Sync context manager representing an opened WebSocket session.
@@ -82,6 +88,7 @@ class WebSocketSession:
 
         self._ping_manager = PingManager()
         self._should_close = threading.Event()
+        self._write_lock: threading.Lock = threading.Lock()
         self._should_close_task: typing.Optional[concurrent.futures.Future[bool]] = None
         self._executor: typing.Optional[concurrent.futures.ThreadPoolExecutor] = None
 
@@ -92,7 +99,7 @@ class WebSocketSession:
 
     def _get_executor_should_close_task(
         self,
-    ) -> typing.Tuple[
+    ) -> tuple[
         concurrent.futures.ThreadPoolExecutor, "concurrent.futures.Future[bool]"
     ]:
         if self._should_close_task is None:
@@ -175,7 +182,8 @@ class WebSocketSession:
         """
         try:
             data = self.connection.send(event)
-            self.stream.write(data)
+            with self._write_lock:
+                self.stream.write(data)
         except httpcore.WriteError as e:
             self.close(CloseReason.INTERNAL_ERROR, "Stream write error")
             raise WebSocketNetworkError() from e
@@ -448,7 +456,8 @@ class WebSocketSession:
             event = wsproto.events.CloseConnection(code, reason)
             data = self.connection.send(event)
             try:
-                self.stream.write(data)
+                with self._write_lock:
+                    self.stream.write(data)
             except httpcore.WriteError:
                 pass
         self.stream.close()
@@ -470,12 +479,13 @@ class WebSocketSession:
         partial_message_buffer: typing.Union[str, bytes, None] = None
         try:
             while not self._should_close.is_set():
-                data = self._wait_until_closed(self.stream.read, max_bytes)
+                data = self._wait_until_closed(self._read_stream, max_bytes)
                 self.connection.receive_data(data)
                 for event in self.connection.events():
                     if isinstance(event, wsproto.events.Ping):
                         data = self.connection.send(event.response())
-                        self.stream.write(data)
+                        with self._write_lock:
+                            self.stream.write(data)
                         continue
                     if isinstance(event, wsproto.events.Pong):
                         self._ping_manager.ack(event.payload)
@@ -502,7 +512,7 @@ class WebSocketSession:
                             self._events.put(full_message_event)
                         continue
                     self._events.put(event)
-        except (httpcore.ReadError, httpcore.WriteError):
+        except (httpcore.ReadError, httpcore.WriteError, EndOfStream):
             self.close(CloseReason.INTERNAL_ERROR, "Stream error")
             self._events.put(WebSocketNetworkError())
         except ShouldClose:
@@ -548,6 +558,12 @@ class WebSocketSession:
             result = todo_task.result()
         return result
 
+    def _read_stream(self, max_bytes: int) -> bytes:
+        data = self.stream.read(max_bytes)
+        if data == b"":
+            raise EndOfStream()
+        return data
+
 
 class AsyncWebSocketSession:
     """
@@ -593,6 +609,7 @@ class AsyncWebSocketSession:
 
         self._ping_manager = AsyncPingManager()
         self._should_close = anyio.Event()
+        self._write_lock = anyio.Lock()
 
         self._max_message_size_bytes = max_message_size_bytes
         self._queue_size = queue_size
@@ -634,7 +651,7 @@ class AsyncWebSocketSession:
 
     async def __aexit__(
         self,
-        exc_type: typing.Optional[typing.Type[BaseException]],
+        exc_type: typing.Optional[type[BaseException]],
         exc: typing.Optional[BaseException],
         tb: typing.Optional[TracebackType],
     ) -> None:
@@ -689,7 +706,8 @@ class AsyncWebSocketSession:
         """
         try:
             data = self.connection.send(event)
-            await self.stream.write(data)
+            async with self._write_lock:
+                await self.stream.write(data)
         except httpcore.WriteError as e:
             await self.close(CloseReason.INTERNAL_ERROR, "Stream write error")
             raise WebSocketNetworkError() from e
@@ -963,7 +981,8 @@ class AsyncWebSocketSession:
             event = wsproto.events.CloseConnection(code, reason)
             data = self.connection.send(event)
             try:
-                await self.stream.write(data)
+                async with self._write_lock:
+                    await self.stream.write(data)
             except httpcore.WriteError:
                 pass
         await self.stream.aclose()
@@ -985,12 +1004,13 @@ class AsyncWebSocketSession:
         partial_message_buffer: typing.Union[str, bytes, None] = None
         try:
             while not self._should_close.is_set():
-                data = await self.stream.read(max_bytes=max_bytes)
+                data = await self._read_stream(max_bytes)
                 self.connection.receive_data(data)
                 for event in self.connection.events():
                     if isinstance(event, wsproto.events.Ping):
                         data = self.connection.send(event.response())
-                        await self.stream.write(data)
+                        async with self._write_lock:
+                            await self.stream.write(data)
                         continue
                     if isinstance(event, wsproto.events.Pong):
                         self._ping_manager.ack(event.payload)
@@ -1017,7 +1037,7 @@ class AsyncWebSocketSession:
                             await self._send_event.send(full_message_event)
                         continue
                     await self._send_event.send(event)
-        except (httpcore.ReadError, httpcore.WriteError):
+        except (httpcore.ReadError, httpcore.WriteError, EndOfStream):
             await self.close(CloseReason.INTERNAL_ERROR, "Stream error")
             await self._send_event.send(WebSocketNetworkError())
 
@@ -1037,10 +1057,16 @@ class AsyncWebSocketSession:
                     )
                     await self._send_event.send(WebSocketNetworkError())
 
+    async def _read_stream(self, max_bytes: int) -> bytes:
+        data = await self.stream.read(max_bytes)
+        if data == b"":
+            raise EndOfStream()
+        return data
+
 
 def _get_headers(
-    subprotocols: typing.Optional[typing.List[str]],
-) -> typing.Dict[str, typing.Any]:
+    subprotocols: typing.Optional[list[str]],
+) -> dict[str, typing.Any]:
     headers = {
         "connection": "upgrade",
         "upgrade": "websocket",
@@ -1065,9 +1091,10 @@ def _connect_ws(
     keepalive_ping_timeout_seconds: typing.Optional[
         float
     ] = DEFAULT_KEEPALIVE_PING_TIMEOUT_SECONDS,
-    subprotocols: typing.Optional[typing.List[str]] = None,
+    subprotocols: typing.Optional[list[str]] = None,
+    session_class: type[SyncSession] = WebSocketSession,  # type: ignore[assignment]
     **kwargs: typing.Any,
-) -> typing.Generator[WebSocketSession, None, None]:
+) -> typing.Generator[SyncSession, None, None]:
     headers = kwargs.pop("headers", {})
     headers.update(_get_headers(subprotocols))
 
@@ -1075,14 +1102,15 @@ def _connect_ws(
         if response.status_code != 101:
             raise WebSocketUpgradeError(response)
 
-        with WebSocketSession(
+        session = session_class(
             response.extensions["network_stream"],
             max_message_size_bytes=max_message_size_bytes,
             queue_size=queue_size,
             keepalive_ping_interval_seconds=keepalive_ping_interval_seconds,
             keepalive_ping_timeout_seconds=keepalive_ping_timeout_seconds,
             response=response,
-        ) as session:
+        )
+        with session:
             yield session
 
 
@@ -1099,9 +1127,10 @@ def connect_ws(
     keepalive_ping_timeout_seconds: typing.Optional[
         float
     ] = DEFAULT_KEEPALIVE_PING_TIMEOUT_SECONDS,
-    subprotocols: typing.Optional[typing.List[str]] = None,
+    subprotocols: typing.Optional[list[str]] = None,
+    session_class: type[SyncSession] = WebSocketSession,  # type: ignore[assignment]
     **kwargs: typing.Any,
-) -> typing.Generator[WebSocketSession, None, None]:
+) -> typing.Generator[SyncSession, None, None]:
     """
     Start a sync WebSocket session.
 
@@ -1134,6 +1163,9 @@ def connect_ws(
             Defaults to 20 seconds.
         subprotocols:
             Optional list of suprotocols to negotiate with the server.
+        session_class:
+            The session class to use.
+            Defaults to [WebSocketSession][httpx_ws.WebSocketSession].
         **kwargs:
             Additional keyword arguments that will be passed to
             the [HTTPX stream()](https://www.python-httpx.org/api/#request) method.
@@ -1168,6 +1200,7 @@ def connect_ws(
                 keepalive_ping_interval_seconds=keepalive_ping_interval_seconds,
                 keepalive_ping_timeout_seconds=keepalive_ping_timeout_seconds,
                 subprotocols=subprotocols,
+                session_class=session_class,
                 **kwargs,
             ) as websocket:
                 yield websocket
@@ -1180,6 +1213,7 @@ def connect_ws(
             keepalive_ping_interval_seconds=keepalive_ping_interval_seconds,
             keepalive_ping_timeout_seconds=keepalive_ping_timeout_seconds,
             subprotocols=subprotocols,
+            session_class=session_class,
             **kwargs,
         ) as websocket:
             yield websocket
@@ -1198,9 +1232,10 @@ async def _aconnect_ws(
     keepalive_ping_timeout_seconds: typing.Optional[
         float
     ] = DEFAULT_KEEPALIVE_PING_TIMEOUT_SECONDS,
-    subprotocols: typing.Optional[typing.List[str]] = None,
+    subprotocols: typing.Optional[list[str]] = None,
+    session_class: type[AsyncSession] = AsyncWebSocketSession,  # type: ignore[assignment]
     **kwargs: typing.Any,
-) -> typing.AsyncGenerator[AsyncWebSocketSession, None]:
+) -> typing.AsyncGenerator[AsyncSession, None]:
     headers = kwargs.pop("headers", {})
     headers.update(_get_headers(subprotocols))
 
@@ -1208,14 +1243,15 @@ async def _aconnect_ws(
         if response.status_code != 101:
             raise WebSocketUpgradeError(response)
 
-        async with AsyncWebSocketSession(
+        session = session_class(
             response.extensions["network_stream"],
             max_message_size_bytes=max_message_size_bytes,
             queue_size=queue_size,
             keepalive_ping_interval_seconds=keepalive_ping_interval_seconds,
             keepalive_ping_timeout_seconds=keepalive_ping_timeout_seconds,
             response=response,
-        ) as session:
+        )
+        async with session:
             yield session
 
 
@@ -1232,9 +1268,10 @@ async def aconnect_ws(
     keepalive_ping_timeout_seconds: typing.Optional[
         float
     ] = DEFAULT_KEEPALIVE_PING_TIMEOUT_SECONDS,
-    subprotocols: typing.Optional[typing.List[str]] = None,
+    subprotocols: typing.Optional[list[str]] = None,
+    session_class: type[AsyncSession] = AsyncWebSocketSession,  # type: ignore[assignment]
     **kwargs: typing.Any,
-) -> typing.AsyncGenerator[AsyncWebSocketSession, None]:
+) -> typing.AsyncGenerator[AsyncSession, None]:
     """
     Start an async WebSocket session.
 
@@ -1267,6 +1304,9 @@ async def aconnect_ws(
             Defaults to 20 seconds.
         subprotocols:
             Optional list of suprotocols to negotiate with the server.
+        session_class:
+            The session class to use.
+            Defaults to [AsyncWebSocketSession][httpx_ws.AsyncWebSocketSession].
         **kwargs:
             Additional keyword arguments that will be passed to
             the [HTTPX stream()](https://www.python-httpx.org/api/#request) method.
@@ -1301,6 +1341,7 @@ async def aconnect_ws(
                 keepalive_ping_interval_seconds=keepalive_ping_interval_seconds,
                 keepalive_ping_timeout_seconds=keepalive_ping_timeout_seconds,
                 subprotocols=subprotocols,
+                session_class=session_class,
                 **kwargs,
             ) as websocket:
                 yield websocket
@@ -1313,6 +1354,7 @@ async def aconnect_ws(
             keepalive_ping_interval_seconds=keepalive_ping_interval_seconds,
             keepalive_ping_timeout_seconds=keepalive_ping_timeout_seconds,
             subprotocols=subprotocols,
+            session_class=session_class,
             **kwargs,
         ) as websocket:
             yield websocket
