@@ -1,7 +1,6 @@
 import contextlib
 import queue
 import typing
-from concurrent.futures import Future
 
 import anyio
 import wsproto
@@ -11,8 +10,8 @@ from wsproto.frame_protocol import CloseReason
 
 from ._exceptions import WebSocketDisconnect
 
-Scope = typing.Dict[str, typing.Any]
-Message = typing.Dict[str, typing.Any]
+Scope = dict[str, typing.Any]
+Message = dict[str, typing.Any]
 Receive = typing.Callable[[], typing.Awaitable[Message]]
 Send = typing.Callable[[Scope], typing.Coroutine[None, None, None]]
 ASGIApp = typing.Callable[[Scope, Receive, Send], typing.Coroutine[None, None, None]]
@@ -40,28 +39,39 @@ class ASGIWebSocketAsyncNetworkStream(AsyncNetworkStream):
         self._send_queue: queue.Queue[Message] = queue.Queue()
         self.connection = wsproto.WSConnection(wsproto.ConnectionType.SERVER)
         self.connection.initiate_upgrade_connection(scope["headers"], scope["path"])
+        self._exit_stack = contextlib.AsyncExitStack()
+        self._aentered = False
 
     async def __aenter__(
         self,
-    ) -> typing.Tuple["ASGIWebSocketAsyncNetworkStream", bytes]:
-        self.exit_stack = contextlib.ExitStack()
-        self.portal = self.exit_stack.enter_context(
-            anyio.from_thread.start_blocking_portal("asyncio")
-        )
-        _: Future[None] = self.portal.start_task_soon(self._run)
+    ) -> tuple["ASGIWebSocketAsyncNetworkStream", bytes]:
+        if self._aentered:
+            raise RuntimeError(
+                "Cannot use ASGIWebSocketAsyncNetworkStream in a context manager twice"
+            )
+        self._aentered = True
+        async with contextlib.AsyncExitStack() as stack:
+            self._task_group = await stack.enter_async_context(
+                anyio.create_task_group()
+            )
+            self._task_group.start_soon(self._run)
 
-        await self.send({"type": "websocket.connect"})
-        message = await self.receive()
+            await self.send({"type": "websocket.connect"})
+            message = await self.receive()
 
-        if message["type"] == "websocket.close":
-            await self.aclose()
-            raise WebSocketDisconnect(message["code"], message.get("reason"))
+            stack.push_async_callback(self.aclose)
 
-        assert message["type"] == "websocket.accept"
-        return self, self._build_accept_response(message)
+            if message["type"] == "websocket.close":
+                await stack.aclose()
+                raise WebSocketDisconnect(message["code"], message.get("reason"))
 
-    async def __aexit__(self, *args: typing.Any) -> None:
-        await self.aclose()
+            assert message["type"] == "websocket.accept"
+            retval = self, self._build_accept_response(message)
+            self._exit_stack = stack.pop_all()
+        return retval
+
+    async def __aexit__(self, exc_type, exc_val, exc_tb) -> typing.Union[bool, None]:
+        return await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)
 
     async def read(
         self, max_bytes: int, timeout: typing.Optional[float] = None
@@ -109,7 +119,6 @@ class ASGIWebSocketAsyncNetworkStream(AsyncNetworkStream):
 
     async def aclose(self) -> None:
         await self.send({"type": "websocket.close"})
-        self.exit_stack.close()
 
     async def send(self, message: Message) -> None:
         self._receive_queue.put(message)
@@ -165,7 +174,7 @@ class ASGIWebSocketTransport(ASGITransport):
         headers = request.headers
 
         if scheme in {"ws", "wss"} or headers.get("upgrade") == "websocket":
-            subprotocols: typing.List[str] = []
+            subprotocols: list[str] = []
             if (
                 subprotocols_header := headers.get("sec-websocket-protocol")
             ) is not None:
@@ -202,7 +211,7 @@ class ASGIWebSocketTransport(ASGITransport):
 
         accept_response_lines = accept_response.decode("utf-8").splitlines()
         headers = [
-            typing.cast(typing.Tuple[str, str], line.split(": ", 1))
+            typing.cast(tuple[str, str], line.split(": ", 1))
             for line in accept_response_lines[1:]
             if line.strip() != ""
         ]
